﻿from __future__ import print_function, absolute_import, division
from KratosMultiphysics import Parameters
from KratosMultiphysics import Vector
from KratosMultiphysics import Matrix

import KratosMultiphysics.KratosUnittest as KratosUnittest

import sys


# input string with ugly formatting
json_string = """
{
   "int_value" : 10,   "double_value": 2.0,   "bool_value" : true,   "string_value" : "hello",
   "level1":
   {
     "list_value":[ 3, "hi", false],
     "tmp" : 5.0
   }
}
"""

pretty_out = """{
    "int_value": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": {
        "list_value": [
            3,
            "hi",
            false
        ],
        "tmp": 5.0
    }
}"""

pretty_out_after_change = """{
    "int_value": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": {
        "list_value": [
            "changed",
            "hi",
            false
        ],
        "tmp": 5.0
    }
}"""

# here the level1 var is set to a double so that a validation error should be thrown
wrong_type = """{
    "int_value": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": 0.0
}"""

# int value is badly spelt
wrong_spelling = """{
    "int_values": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": 0.0
}"""

# wrong on the first level
# error shall be only detective by recursive validation
wrong_lev2 = """{
    "int_value": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": { "a":0.0 }
}"""

defaults = """
{
	"int_value": 10,
	"double_value": 2.0,
	"bool_value": false,
	"string_value": "hello",
	"level1": {
		"list_value": [
			3,
			"hi",
			false
		],
		"tmp": "here we expect a string"
	},
	"new_default_value": -123.0,
	"new_default_obj": {
		"aaa": "string",
		"bbb": false,
		"ccc": 22
	}
}
"""


expected_validation_output = """{
    "int_value": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": {
        "list_value": [
            3,
            "hi",
            false
        ],
        "tmp": 5.0
    },
    "new_default_value": -123.0,
    "new_default_obj": {
        "aaa": "string",
        "bbb": false,
        "ccc": 22
    }
}"""

four_levels = """{
    "int_value": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": { 
        "level2": {
            "level3": {
                "level4": {
                }
            }
        } 
    }
}"""

four_levels_variation = """{
    "int_value": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": { 
        "a":11.0,
        "level2": {
            "level3": {
                "level4": {
                }
            }
        } 
    }
}"""

four_levels_wrong_variation = """{
    "int_value": 10,
    "double_value": "hi",
    "bool_value": true,
    "string_value": "hello",
    "level1": { 
        "a":11.0,
        "level2": {
            "level3": {
                "level4": {
                }
            }
        } 
    }
}"""

four_levels_defaults = """{
    "int_value": 10,
    "double_value": 2.0,
    "bool_value": true,
    "string_value": "hello",
    "level1": { 
        "a":1.0,
        "level2": {
            "b":2.0,
            "level3": {
                "c":3.0,
                "level4": {
                    "d":4.0
                }
            }
        } 
    }
}"""

class TestParameters(KratosUnittest.TestCase):    

    def setUp(self):
        self.kp = Parameters(json_string)
        self.compact_expected_output = """{"int_value":10,"double_value":2.0,"bool_value":true,"string_value":"hello","level1":{"list_value":[3,"hi",false],"tmp":5.0}}"""
        
        if (sys.version_info < (3, 2)):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_kratos_parameters(self):
        self.assertEqual(
            self.kp.WriteJsonString(),
            self.compact_expected_output
        )

        self.assertTrue(self.kp.Has("int_value"))
        self.assertFalse(self.kp.Has("unextisting_value"))

        self.assertEqual(self.kp["int_value"].GetInt(), 10)
        self.assertEqual(self.kp["double_value"].GetDouble(), 2.0)
        self.assertEqual(self.kp["bool_value"].GetBool(), True)
        self.assertEqual(self.kp["string_value"].GetString(), "hello")

        self.assertEqual(self.kp.PrettyPrintJsonString(), pretty_out)

    def test_kratos_change_parameters(self):
        # now change one item in the sublist
        subparams = self.kp["level1"]

        my_list = subparams["list_value"]

        for i in range(my_list.size()):
            if my_list[i].IsBool():
                self.assertEqual(my_list[i].GetBool(), False)

        # my_list = subparams["list_value"]
        subparams["list_value"][0].SetString("changed")

        self.assertEqual(
            self.kp.PrettyPrintJsonString(),
            pretty_out_after_change
        )

    def test_kratos_copy_parameters(self):
        # try to make a copy
        original_out = self.kp.PrettyPrintJsonString()
        other_copy = self.kp.Clone()

        self.assertEqual(
            other_copy.PrettyPrintJsonString(),
            original_out
        )

        other_copy["int_value"].SetInt(-1)
        self.assertEqual(self.kp["int_value"].GetInt(), 10)
        # self.assertEqual(other_copy["int_value").GetString(),-1)

    def test_set_value(self):
        kp = Parameters(json_string)
        kp1 = Parameters(pretty_out_after_change)

        kp["bool_value"] = kp1["level1"]
        kp["bool_value"].PrettyPrintJsonString()
        self.assertEqual(kp["bool_value"].PrettyPrintJsonString(), kp1["level1"].PrettyPrintJsonString())

    def test_kratos_wrong_parameters(self):
        # should check which errors are thrown!!
        with self.assertRaisesRegex(RuntimeError, "no_value"):
            self.kp["no_value"].GetInt()

    def test_validation_fails_due_to_wrong_type(self):
        kp = Parameters(wrong_type)
        defaults_params = Parameters(defaults)

        # should check which errors are thrown!!
        with self.assertRaises(RuntimeError):
            kp.ValidateAndAssignDefaults(defaults_params)

    def test_validation_fails_due_to_wrong_spelling(self):
        kp = Parameters(wrong_spelling)
        defaults_params = Parameters(defaults)

        # should check which errors are thrown!!
        with self.assertRaises(RuntimeError):
            kp.ValidateAndAssignDefaults(defaults_params)

    def test_recursive_validation_fails_error_on_first_level(self):
        kp = Parameters(wrong_lev2)
        defaults_params = Parameters(defaults)

        # should check which errors are thrown!!
        with self.assertRaises(RuntimeError):
            kp.RecursivelyValidateAndAssignDefaults(defaults_params)

    def test_recursive_validation_4_levels(self):
        kp = Parameters(four_levels)
        kp_variation = Parameters(four_levels_variation)
        kp_wrong_wariation = Parameters(four_levels_wrong_variation)
        defaults_params = Parameters(four_levels_defaults)

        kp.RecursivelyValidateAndAssignDefaults(defaults_params)
        kp_variation.RecursivelyValidateAndAssignDefaults(defaults_params)

        self.assertTrue( kp.IsEquivalentTo(defaults_params) )
        self.assertFalse( kp_variation.IsEquivalentTo(defaults_params) )
        
        self.assertTrue( kp.HasSameKeysAndTypeOfValuesAs(defaults_params) )
        self.assertTrue( kp_variation.HasSameKeysAndTypeOfValuesAs(defaults_params) )
        self.assertFalse( kp_wrong_wariation.HasSameKeysAndTypeOfValuesAs(defaults_params) )
        
    def test_validation_succeds_error_on_first_level(self):
        kp = Parameters(wrong_lev2)
        defaults_params = Parameters(defaults)

        # here no error shall be thrown since validation is only done on level0
        kp.ValidateAndAssignDefaults(defaults_params)

    def test_validation_succeeds(self):
        kp = Parameters(json_string)
        defaults_params = Parameters(defaults)
        defaults_params["level1"]["tmp"].SetDouble(2.0)  # this does not coincide with the value in kp, but is of the same type

        kp.ValidateAndAssignDefaults(defaults_params)
        self.assertEqual(kp.PrettyPrintJsonString(), expected_validation_output)

        self.assertEqual(kp["level1"]["tmp"].GetDouble(), 5.0)  # not 2, since kp overwrites the defaults
        
    def test_add_value(self):
        kp = Parameters("{}")
        kp.AddEmptyValue("new_double").SetDouble(1.0)
        
        self.assertTrue(kp.Has("new_double"))
        self.assertEqual(kp["new_double"].GetDouble(), 1.0)
        
    def test_iterators(self):
        kp = Parameters(json_string)
        
        #iteration by range
        nitems = 0
        for iterator in kp:
            nitems = nitems + 1
        self.assertEqual(nitems, 5)
            
        #iteration by items
        for key,value in kp.items():
            #print(value.PrettyPrintJsonString())
            self.assertEqual(kp[key].PrettyPrintJsonString(), value.PrettyPrintJsonString())
            #print(key,value)
            
        #testing values
        expected_values = ['10', '2.0', 'true', '"hello"', '{"list_value":[3,"hi",false],"tmp":5.0}']
        counter = 0
        
        for value in kp.values():
            self.assertEqual(value.WriteJsonString(), expected_values[counter])
            counter += 1

        #testing values
        expected_keys = ['int_value', 'double_value', 'bool_value', 'string_value', 'level1'] 
        counter = 0
        for key in kp.keys():
            self.assertEqual(key, expected_keys[counter])
            counter += 1 

    def test_remove_value(self):
        kp = Parameters(json_string)
        self.assertTrue(kp.Has("int_value"))
        self.assertTrue(kp.Has("level1"))
                         
        kp.RemoveValue("int_value")
        kp.RemoveValue("level1")
        
        self.assertFalse(kp.Has("int_value"))
        self.assertFalse(kp.Has("level1"))

    def test_is_methods(self):
        # This method checks all the "IsXXX" Methods
        tmp = Parameters("""{
            "int_value" : 10,   
            "double_value": 2.0,   
            "bool_value" : true,   
            "string_value" : "hello",
            "vector_value" : [5,3,4], 
            "matrix_value" : [[1,2],[3,6]]
        }""") # if you add more values to this, make sure to add the corresponding in the loop

        for key in tmp.keys():
            val_type = key[:-6] # removing "_value"

            if val_type == "int":
                self.assertTrue(tmp[key].IsInt())
            else:                    
                self.assertFalse(tmp[key].IsInt())

            if val_type == "double":
                self.assertTrue(tmp[key].IsDouble())
            else:
                self.assertFalse(tmp[key].IsDouble())

            if val_type == "bool":
                self.assertTrue(tmp[key].IsBool())
            else:
                self.assertFalse(tmp[key].IsBool())

            if val_type == "string":
                self.assertTrue(tmp[key].IsString())
            else:
                self.assertFalse(tmp[key].IsString())

            if val_type == "vector":
                self.assertTrue(tmp[key].IsVector())
            else:
                self.assertFalse(tmp[key].IsVector())

            if val_type == "matrix":
                self.assertTrue(tmp[key].IsMatrix())
            else:
                self.assertFalse(tmp[key].IsMatrix())   

    def test_get_methods(self):
        # This method checks all the "GetXXX" Methods if they throw an error
        tmp = Parameters("""{
            "int_value" : 10,   
            "double_value": 2.0,   
            "bool_value" : true,   
            "string_value" : "hello",
            "vector_value" : [5.2,-3.1,4.33], 
            "matrix_value" : [[1,2],[3,4],[5,6]]
        }""") # if you add more values to this, make sure to add the corresponding in the loop

        for key in tmp.keys():
            val_type = key[:-6] # removing "_value"

            # Int and Double are checked tgth bcs both internally call "IsNumber"
            if val_type == "int" or val_type == "double": 
                if val_type == "int":
                    self.assertEqual(tmp[key].GetInt(),10)
            else:
                with self.assertRaises(RuntimeError):
                    tmp[key].GetInt()
            
            if val_type == "double" or val_type == "int":
                if val_type == "double":
                    self.assertEqual(tmp[key].GetDouble(),2.0)
            else:
                with self.assertRaises(RuntimeError):
                    tmp[key].GetDouble()
            
            if val_type == "bool":
                self.assertEqual(tmp[key].GetBool(),True)
            else:
                with self.assertRaises(RuntimeError):
                    tmp[key].GetBool()
            
            if val_type == "string":
                self.assertEqual(tmp[key].GetString(),"hello")
            else:
                with self.assertRaises(RuntimeError):
                    tmp[key].GetString()

            if val_type == "vector":
                V = tmp[key].GetVector()
                self.assertEqual(V[0],5.2)
                self.assertEqual(V[1],-3.1)
                self.assertEqual(V[2],4.33)
            else:
                with self.assertRaises(RuntimeError):
                    tmp[key].GetVector()

            if val_type == "matrix":
                A = tmp[key].GetMatrix()
                self.assertEqual(A[0,0],1.0)
                self.assertEqual(A[0,1],2.0)
                self.assertEqual(A[1,0],3.0)
                self.assertEqual(A[1,1],4.0)
                self.assertEqual(A[2,0],5.0)
                self.assertEqual(A[2,1],6.0)
            else:
                with self.assertRaises(RuntimeError):
                    tmp[key].GetMatrix()   
        
    def test_vector_interface(self):
        tmp = Parameters("""{ }""")
        # Manually assign and check a Vector
        vec = Vector(3)
        vec[0] = 1.32
        vec[1] = -2.22
        vec[2] = 5.5

        tmp.AddEmptyValue("vector_value")
        tmp["vector_value"].SetVector(vec)

        self.assertTrue(tmp["vector_value"].IsVector())

        V2 = tmp["vector_value"].GetVector()
        self.assertEqual(V2[0],1.32)
        self.assertEqual(V2[1],-2.22)
        self.assertEqual(V2[2],5.5)

    def test_matrix_interface(self):
        # Read and check a Matrix from a Parameters-Object
        tmp = Parameters("""{
            "false_matrix_value": [[2, 1.5,3.3],[1,2]]
        }""")
        
        # Manually assign and check a Matrix
        mat = Matrix(3,2)
        mat[0,0] = 1.0
        mat[0,1] = 2.0
        mat[1,0] = 3.0
        mat[1,1] = 4.0
        mat[2,0] = 5.0
        mat[2,1] = 6.0

        tmp.AddEmptyValue("matrix_value")
        tmp["matrix_value"].SetMatrix(mat)
        
        self.assertTrue(tmp["matrix_value"].IsMatrix())
        
        A2 = tmp["matrix_value"].GetMatrix()
        self.assertEqual(A2[0,0],1.0)
        self.assertEqual(A2[0,1],2.0)
        self.assertEqual(A2[1,0],3.0)
        self.assertEqual(A2[1,1],4.0)
        self.assertEqual(A2[2,0],5.0)
        self.assertEqual(A2[2,1],6.0)

        # Check the IsMatrix Method
        self.assertFalse(tmp["false_matrix_value"].IsMatrix()) # Mis-sized Matrix

        # check that the errors of the GetMatrix method are thrown correctly
        with self.assertRaises(RuntimeError):        
            tmp["false_matrix_value"].GetMatrix() # Mis-sized Matrix
        
        
if __name__ == '__main__':
    KratosUnittest.main()
